classdef SignalAnalysis
    %SIGNALANALYSIS Summary of this class goes here
    %   Detailed explanation goes here
    
    properties (Access = public)       
    end
    methods(Static)    
        function pruebaFinal = realizarPruebasAcusticas(prueba, audio, gender, mpt, fs)
            arguments              
                prueba
                audio (:,1) double
                gender
                mpt (1,1) double {mustBeFinite} = -1;
                fs (1,1) double {mustBeFinite} = 96e3;
            end
            audio = SignalAnalysis.normalizaAudioPordB(audio, -20);
            prueba.jitter = SignalAnalysis.calcular_Jitter(audio);
            prueba.shimmer = SignalAnalysis.calcular_Shimmer(audio); 
            prueba.ilow = SignalAnalysis.calculate_min_intensity(audio);
            prueba.snr = SignalAnalysis.calculate_snr(audio);
            prueba.nne = SignalAnalysis.calculate_nne(audio, prueba.snr);
            prueba.hnr = SignalAnalysis.calculate_hnr(audio);
            [meanF0, minF0, maxF0] = SignalAnalysis.getFrecuenciaFundamentalAudio(audio);
            prueba.meanF0 = meanF0;
            prueba.minF0 = minF0;
            prueba.maxF0 = maxF0;
            if(mpt == -1)
                mpt = SignalAnalysis.ObtenerTiempoMaximoFonacion(audio);
            end
            prueba.dsi = SignalAnalysis.calcular_Dysphonia_Severity_Index(mpt, prueba.maxF0, prueba.ilow, prueba.jitter);
            formants = SignalAnalysis.calculate_formants_lpc(audio, gender, fs);
            prueba.f1 = formants(1);
            prueba.f2 = formants(2);
            prueba.f3 = formants(3);
            directory = append('./', prueba.iduser, '/Audio/', prueba.idPrueba);
            prueba.narrow_spec = SignalAnalysis.plotAudioSpectrogram(audio, 'Narrow', directory, fs);
            prueba.wide_spec = SignalAnalysis.plotAudioSpectrogram(audio, 'Wide', directory, fs);
            prueba.fft = SignalAnalysis.plotAudioFft(audio, directory, fs);
            %prueba.lpc = SignalAnalysis.plotAudioLpc(audio, directory, fs);
            prueba.cepstrum = SignalAnalysis.plotAudioCepstrum(audio, directory, fs);
            pruebaFinal = prueba;
        end

        function fftSignal = ObtenerFFT(audio)
            fftSignal = fft(audio);
        end

        function pico = ObtenerFrequenciaPico(audio, fs, tipo)
            arguments
                audio (:,1) double
                fs (1,1) double
                tipo (1,:) char {mustBeMember(tipo,{'Maximo','Minimo'})}
            end
            fftSignal = SignalAnalysis.ObtenerFFT(audio);
            L = length(audio); % sample length
            P2 = abs(fftSignal/L);
            P1 = P2(1:L/2+1);
            P1(2:end-1) = 2*P1(2:end-1);
            
            f = fs*(0:(L/2))/L;
            plot(f,P1) 
            title("Single-Sided Amplitude Spectrum of S(t)")
            xlabel("f (Hz)")
            ylabel("|P1(f)|")
            if(tipo == "Maximo")
                pico = max(P1);
            elseif(tipo == "Minimo")
                pico = min(P1);
            end
        end

        function tiempo = ObtenerTiempoMaximoFonacion(audio, fs)
            arguments
                audio (:,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
            end
            N = length(audio); % sample length
            tiempo = N/fs; % total time span of audio signal
        end

        function audioFinal = ObtenerAudioSinHabla(audio, fs)
            arguments
                audio (:,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
            end
%             Detectamos los indices del audio donde se ha detectado habla
            indicesHabla = detectSpeech(audio,fs);
%             Creamos una variable de array vacio
            f0Audio = [];
%             Recorremos todos los indices que se han detectado en el habla
            for ii = 1:size(indicesHabla,1)
%                 Recuperamos desde la parte del audio que coincide con la
%                 parte hablada
                segmentoHablado = audio(indicesHabla(ii,1):indicesHabla(ii,2));
%                 Guardamos el pitch recuperado
                f0Audio = [f0Audio;pitch(segmentoHablado,fs)];
            end
            audioFinal = f0Audio;
        end
        
        function audioDetected = ObtenerAudioVoiceDetected(audioFileName, fs)
            arguments
                audioFileName (1,:) char
                fs (1,1) double {mustBeFinite} = 96e3;
            end
            fileReader = dsp.AudioFileReader(audioFileName, "SamplesPerFrame", fs);            
            VAD = voiceActivityDetector;
            audioDetected = [];
            while ~isDone(fileReader)
                x = fileReader();
                
                if VAD(x) > 0.99
%                     decision = pitch(x,fileReader.SampleRate, ...
%                         WindowLength= size(x,1), ...
%                         OverlapLength=0);
                else
                    decision = NaN;
                end
                audioDetected = [audioDetected;decision];
            end            
        end

        function fileName = plotAudioSpectrogram(audio, bandSize, directory, fs)
            arguments                
                audio (:,1) double
                bandSize (1,:) char {mustBeMember(bandSize,{'Narrow','Wide'})}
                directory (1,:) char
                fs (1,1) double {mustBeFinite} = 96e3;
            end

            if(bandSize == "Narrow")
%                 45Hz
                tSize = 0.03;
            else
%                 250Hz
                tSize = 4e-3;
            end

            windowSize = round(tSize * fs);
            % definimos el tamaño del overlap como el 8% de la ventana
            overlap = round(windowSize * 0.08);
            %  generamos el espectrograma
            f1 = figure;%('visible','off');
            spectrogram(audio, windowSize, overlap, fs, fs, 'yaxis')
            colormap jet
            ylim([0 8])
            xlabel('Tiempo (s)')
            ylabel('Frecuencia (kHz)')
            %set(gcf, 'Position', [2500 100 1550 800]);
            nombre = append('espectograma', Util.GenerarUUID());
            saveas(f1, append(directory, '/', nombre),'png')
            fileName = nombre;
        end

        function fileName = plotAudioFft(audio, directory, fs)
            arguments                
                audio (:,1) double
                directory (1,:) char
                fs (1,1) double {mustBeFinite} = 96e3;
            end
            lenX = length(audio);
            
            %calculamos el tamaño de la fft para evitar aliasing
            nfft = 2^nextpow2(lenX);

            %calculamos las frecuencias de la fft
            freqFft = (0 : (nfft/2 - 1)) * (fs/nfft);
            Xs = fft(audio, nfft);
           
            nombre = append('freqPositivas', Util.GenerarUUID());
            %plot de las frecuencias positivas
            f2 = figure;%('visible','off');
            plot(freqFft, abs(Xs(1:nfft/2)))
            xlabel('f(Hz)')
            ylabel('|X(f)|')
            xlim([0 5000])
            ylim([0 2000])
            saveas(f2,append(directory, '/', nombre),'png')
            fileName = nombre;
        end

        function fileName = plotAudioCepstrum(audio, directory, fs)
            arguments                
                audio (:,1) double
                directory (1,:) char
                fs (1,1) double {mustBeFinite} = 96e3;
            end
            c = rceps(audio);
            mpt = SignalAnalysis.ObtenerTiempoMaximoFonacion(audio, fs);
            ts = mpt/length(c);
            t = ts: ts : mpt;
            tiempo = SignalAnalysis.ObtenerTiempoMaximoFonacion(audio, fs);
            nombre = append('cepstrum', Util.GenerarUUID());
            f3 = figure;%('visible','off');
            plot(t,c)
            xlabel('Tiempo (s)')
            ylabel('Amplitud')
            xlim([tiempo/2 ((tiempo/2) + 0.03)])
            ylim([-4e-3 4e-3])
            saveas(f3,append(directory, '/', nombre),'png')
            fileName = nombre;
        end

        function audioFiltrado = filtroPasoAlto(audio, fs, fHp)
            arguments
                audio (:,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
                fHp (1,1) double {mustBeFinite} = 50; % Frecuencia de paso del filtro paso alto recomendada(50Hz)
            end
            audioFiltrado = highpass(audio, fHp, fs, 'ImpulseResponse','iir');
        end

        function comprimirAudio(audio, fs, threshold, ratio)
            arguments
                audio (:,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
                threshold (1,1) double {mustBeFinite} = -10;
                ratio (1,1) double {mustBeFinite} = 5;
            end
            %leemos el archivo de audio
            dRC = compressor(threshold, ratio, "SampleRate", fs);
            lenX = length(audio);
            
            audioOut = dRC(audio);
            %calculamos el tamaño de la fft para evitar aliasing
            nfft = 2^nextpow2(lenX);

            %calculamos las frecuencias de la fft
            freqFft = (0 : (nfft/2 - 1)) * (fs/nfft);
            Xs = fft(audio, nfft);
            XsCompressed = fft(audioOut, nfft);
            figure(2)
            plot(freqFft, abs(Xs(1:nfft/2)))
            xlabel('f(Hz)')
            ylabel('|X(f)|')
            xlim([0 5000])
            ylim([0 2000])

            figure(1)
            plot(freqFft, abs(XsCompressed(1:nfft/2)))
            xlabel('f(Hz)')
            ylabel('|X(f)|')
            xlim([0 5000])
            ylim([0 2000])
        end

        function audio_normalizado = normalizaAudioPordB(audio, dB)
            arguments
                audio (:,1) double
                dB (1,1) double {mustBeFinite} = -50;
            end
%             Calculamos el Power rate(PR) en funcion de los dB
            pR = 10^(dB/10);
%             Calculamos el ratio de amplitud en funcion del PR
            A1 = sqrt(pR);
            audio_normalizado = audio * A1 / max( abs( audio(:) ) );
        end

        function [meanF0, minF0, maxF0] = getFrecuenciaFundamentalAudio(audio, fs)
            arguments
                audio (:,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
            end
            wLength = round(fs*0.052);
            audioLength = size(audio,1);
            if wLength > audioLength
                wLength = round(audioLength*0.052);
            end
            if(wLength<1)
                wLength = 1;
            end
            overLapLength = round(fs*0.042);
            if(overLapLength > wLength)
                overLapLength = round(audioLength*0.042);
            end

            f0 = pitch(audio,fs,Method="SRH", ...
                Range=[40, 400], ...
                WindowLength=wLength, ...
                OverlapLength=overLapLength);
            meanF0 = mean(f0, "all", "omitnan");
            f0 = pitch(audio,fs,Method="SRH", ...
                Range=[80, 1100], ...
                WindowLength=wLength, ...
                OverlapLength=overLapLength);
            minF0 = min(f0(f0 >= 50), [],"omitnan");
            maxF0 = max(f0,[],"omitnan");
        end

        function jitter = calcular_Jitter(audio, fs, gp)
             arguments
                audio (:,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
%                 duracion de un ciclo glotal
                gp (1,1) double {mustBeFinite} = 8.3e-3;
             end
%  Calculamos el tamaño de la ventana en funcion del tiempo de un
% ciclo glotico y la frecuencia de muestreo del audio
            L = length(audio);
            wLength = round(L / (gp * fs));
            ovLength = round(wLength * 0.807);
%             Calculamos el pitch del audio
            f0 = pitch(audio,fs,Method="SRH", ...
                WindowLength=wLength, ...
                OverlapLength=ovLength);
%             Calculamos solo con aquellos que sean superiores al 70% de la
%             media
            meanf0 = mean(f0, "all", "omitnan");
            f0 = f0(f0 >= 0.7*meanf0);
            N = length(f0);
            j = 0;
            t = 0;
            for a = 1 : (N-1)
                 j = j + norm((1/f0(a)) - (1/f0(a + 1)));
                 t = t + 1/f0(a);
            end
            t = t + f0(N);
            jitter = ((1 / (N - 1) * j)/((1/N)*t))*100;
        end

        function shimmer = calcular_Shimmer(audio, fs, gp)
             arguments
                audio (:,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
%                 duracion de un ciclo glotal
                gp (1,1) double {mustBeFinite} = 8.3e-3;
             end            
%             Calculamos solo con aquellos que sean superiores al 70% de la
%             media
            meanAmplitude = mean(abs(audio), "all", "omitnan");
            minAmplitude = 0.7 * meanAmplitude;
            sampleLength = round(gp * fs);
            L = length(audio);
            if(mod(L/sampleLength, 1) ~= 0)
                x = ceil(L/sampleLength);
                valoresDeRelleno = (x * sampleLength) - L;
                audio (end + 1 : end + valoresDeRelleno) = NaN;
            end
            R = reshape(audio,sampleLength,[]);
            R = max(R, [], 1);
            R = R(R >= minAmplitude);
            N = length(R);
            s = 0;
            a = 0;
            for i = 1 : (N-1)                              
                s = s + norm(R(i) - R(i + 1));
                a = a + R(i);
            end
            a = a + R(N);
            shimmer = ((1 / (N - 1) * s)/((1/N)*a))*100;
        end

        function dsi = calcular_Dysphonia_Severity_Index(mpt, f0High, iLow, jitter)
        % calcular_Dysphonia_Severity_Index dados los parametros calcula el
        % Dysphonia Severity Index (DSI)
        % mpt: Maximo tiempo de fonación (s)
        % f0High: Frecuencia fundamental máxima (Hz)
        % iLow: intensidad más baja (ILow) del Rango de Voz (dB)
        % jitter: Jitter relativo (%)
            arguments
                mpt (1,1) double
                f0High (1,1) double
                iLow (1,1) double
                jitter (1,1) double
            end   
            dsi = 0.13 * mpt + 0.0053 * f0High - 0.26 * iLow - 1.18 * jitter + 12.4;
        end

        function audiodB = pasar_Audio_dB(audio)
        % calcular_Dysphonia_Severity_Index dados los parametros calcula el
        % Dysphonia Severity Index (DSI)
        % mpt: Maximo tiempo de fonación (s)
        % f0High: Frecuencia fundamental máxima (Hz)
        % iLow: intensidad más baja (ILow) del Rango de Voz (dB)
        % jitter: Jitter relativo (%)
            arguments
                audio (:,1) double
            end   
            audiodB = (10*log10(audio.^2)) + 90.308998699194362;
        end

        function ilow = calculate_min_intensity(audio)
             arguments
                audio (:,1) double
             end 
             audiodB = SignalAnalysis.pasar_Audio_dB(audio);
             audiodB(isinf(audiodB) | isnan(audiodB) | audiodB == 0) = NaN;
%              meandB = mean(abs(audiodB), "all", "omitnan");
%              mindB = 0.1 * meandB;
%              audiodB = audiodB(audiodB >= mindB);
             ilow = min(audiodB(audiodB > 0), [],"omitnan");
        end

        function snr = calculate_snr(audio)
            % calculate_snr: Calcula el Signal-to-noise ratio (SNR) de la
            % señal de voz haciendo uso de la desviación estandar puesto
            % que no se posee una grabacion aislada del ruido
            arguments
                audio (:,1) double
            end 
            % Obtenemos los valores de intensidad del audio
            audiodB = SignalAnalysis.pasar_Audio_dB(audio);
            audiodB(isinf(audiodB) | isnan(audiodB) | audiodB == 0) = NaN;

            % Obtenemos la media
            u = mean(audiodB, "omitnan");

            % obtenemos la desviacion estandar
            sd = std(audiodB,"omitnan");

            % calculamos el SNR
            snr = u / sd;
        end

        function nne = calculate_nne(audio, snr, fs)
            arguments
                audio (:,1) double
                snr (1,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
            end 
            % Obtenemos la potencia del audio
            pRMS = rms(audio)^2;
            powbp = bandpower(audio, fs, [0 fs/2]);

            % Calculamos el SNR en funcion de los dB
            Psnr = 10^(snr/10);

            Pn = powbp/ (Psnr - 1);
            nne = 20*log10(Pn) + 90.308998699194362;
        end

        function hnr = calculate_hnr(audio)
            arguments
                audio (:,1) double
            end     
            % calculamos el maximo           
            audio = audio - mean(audio);
            maximo = max(abs(audio));
            
            % normalizamos entre -1 y 1
            if maximo <= 0
                hnr = 0;
                return;
            end

            audio = audio / maximo;
            
            % Calculamos la autocorrelacion y mantenemos los indices
            % positivos
            corr = xcorr(audio, audio);
            corr = corr((length(corr)/2) + 1 : end); 
            
            % Buscamos los picos
            [~, locs] = findpeaks(corr); 
            
            if isempty(locs)
                hnr = 0;
                return;
            end

            % localizamos el primer pico
            primer_pico = locs(1);

            % obtenemos el valor correspondiente a T0
            [~, T0] = max(corr(primer_pico:end));
            T0 = T0 + primer_pico;
            
            %calculamos el HNR
            hnr = 10.0 * log10(corr(T0) / (max(corr) - corr(T0)));
        end

        function formants = calculate_formants_lpc(audio, gender, fs)
            arguments
                audio (:,1) double
                gender
                fs (1,1) double {mustBeFinite} = 96e3;
            end
            % enventanamos el audio utilizando una ventana de hamming
            L = length(audio);
            xn = audio.*hamming(L);
            
            %aplicamos filtro de preenfasis
            preemph = [1 0.63];
            xn = filter(1,preemph,xn);

            % calculamos el numero de coeficientes que se estima que se van 
            % a encontrar
            nCoef = SignalAnalysis.calculate_number_coefficients(gender, fs);

            %calculamos las raices del polinomio que devuelve lpc
            A = lpc(xn, nCoef);
            rts = roots(A);

            % determinamos los angulos de las raices
            rts = rts(imag(rts)>=0);
            angz = atan2(imag(rts),real(rts));
            
            [frqs,locs] = sort(angz.*(fs/(2*pi)));
            bw = -1/2*(fs/(2*pi))*log(abs(rts(locs)));
           
            n = 1;
            for i = 1:length(frqs)
                if (frqs(i) > 90 && bw(i) < 400)
                    formants(n) = frqs(i);
                    n = n+1;
                end
            end            
        end

        function nCoef = calculate_number_coefficients(gender, fs)
            arguments
                gender
                fs (1,1) double {mustBeFinite} = 96e3;
            end
            %calculamos segun Gunnar Fant el intervalo en hercios al que se
            % espera que ocurran los formantes
            % C/2L 
            % C: velocidad del sonidoen cm/s. L: longitud del tracto vocal
            % en cm
            C = 34300;
            if (gender == "M")
                L = 17.5;
            else
                L = 12.75;
            end

            interHz = C / (2 * L);
            
            %calculamos el numero de formantes
            nForm = round((fs/2) / interHz);

            %calculamos el numero de coeficientes que se esperan
            nCoef = nForm * 2 + 2;
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %SIN USO
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        function hnr = DEPRECATED_calculate_hnr(audio, fs)
            arguments
                audio (:,1) double
                fs (1,1) double {mustBeFinite} = 96e3;
            end 

            % calculamos la autocorrelacion
            max_lag = length(audio) - 1;
            autocorr_result = xcorr(audio, max_lag, 'coeff');
            
            % Find peaks in autocorrelation (peaks correspond to harmonics)
            [pks, locs] = findpeaks(autocorr_result); 
            
            % Extract harmonic components based on detected peaks
            harmonic_components = zeros(size(audio));
            for i = 1:length(locs)
                harmonic_period = locs(i);
                harmonic_components(harmonic_period:harmonic_period:end) = audio(harmonic_period:harmonic_period:end);
            end
            
            % Calculate energies
            energy_harmonic = sum(harmonic_components.^2);
            energy_residual = sum((audio - harmonic_components).^2);
            
            % Calculate HNR in dB
            hnr = 10 * log10(energy_harmonic / energy_residual);
                    end
            
        function n = DEPRECATED_getHNR(y, F0, Fs, Nfreqs)
            
            NBins = length(y);
            N0 = round(Fs / F0);
            N0_delta = round(N0 * 0.1);  % search 10% either side
            
            y = y .* hamming(length(y));
            Y = fft(y, NBins);
            aY = log10(abs(Y));
            ay = ifft(aY);
            
            % find possible rahmonic peaks
            peakinx = zeros(floor(length(y)/ 2 / N0), 1);
            for k=1:length(peakinx)
                ayseg = ay(k*N0 - N0_delta : k*N0 + N0_delta);
                [val, inx] = max(abs(ayseg));
                peakinx(k) = inx + k*N0 - N0_delta - 1;
            
                % lifter out each rahmonic peak
                s_ayseg = sign(diff(ayseg));
                
                % find first derivative sign change
                l_inx = inx - find((s_ayseg(inx-1:-1:1) ~= sign(inx)), 1) + 1;
                r_inx = inx + find((s_ayseg(inx+1:end) == sign(inx)), 1);
                
                l_inx = l_inx + k*N0 - N0_delta - 1;
                r_inx = r_inx + k*N0 - N0_delta - 1;
                
                % lifter out the peak
                ay(l_inx:r_inx) = 0;
               
            end
            
            % put the signal back together
            midL = round(length(y) / 2) + 1;
            ay(midL : end) = ay(midL - 1: -1 : midL - 1 - (length(ay) - midL));
            
            Nap = real(fft(ay));
            N = Nap;
            Ha = aY - Nap;  % approximated harmonic spectrum
            
            % calculate baseline corrections
            Hdelta = F0 / Fs * length(y);
            for f=Hdelta+0.0001:Hdelta:round(length(y)/2)
                fstart = ceil(f - Hdelta);
                
                Bdf = abs(min(Ha(fstart:round(f))));
                N(fstart:round(f)) = N(fstart:round(f)) - Bdf;
            end
            
            % calculate the average HNR for Nfreqs
            H = aY - N;  %note that N is valid only for 1:length(N)/2
            n = zeros(length(Nfreqs), 1);
            for k=1:length(Nfreqs)
                Ef = round(Nfreqs(k) / Fs * length(y));  % equivalent cut off frequency
                n(k) = 20*mean(H(1:Ef)) - 20*mean(N(1:Ef));
            end
        
        end

    end
end

