classdef Util
    % Clase que recoje las funciones utiles que se van a utilizar en
    % repetidas ocasiones en toda la aplicacion
     
    methods(Static)
        function guardarArchivoEnDirectorio(rutaCliente, extension)
            % Esta funcion permite seleccionar un archivo y guardarlo en
            % la carpeta que corresponda al cliente dentro de la ruta del
            % programa

            % comprobamos que tipo de archivo se va a recoger y asignamos 
            % las extensiones pertinentes para la posterior busqueda  
            if strcmp (extension, 'img')
                extensiones = {'*.jpg; *.jpeg; *.png'};            
            elseif strcmp (extension, 'pdf')
                extensiones = {'*.pdf'};
            end
%             Se pide al usuario seleccionar un archivo para su subida
            [FileName, PathName] = uigetfile(extensiones, 'Seleccione el archivo');
            % Se monta la ruta del archivo
            uploadedFullFileName = fullfile(PathName, FileName);
            % se crea el directorio donde se va a guardar
            status = mkdir(rutaCliente);
            if(status == 1) 
                %extraemos la extension del archivo seleccionado
                [filepath, name, ext] = fileparts(FileName);
                % creamos una uuid para cerciorarnos de que no se repetiran
                % los nombres de los archivos en el directorio
                uuid = char(java.util.UUID.randomUUID);
                fileName = append(uuid, ext);
                fullFileName = fullfile(rutaCliente, fileName);
                % Copiamos el archivo de la direccion de             
                copyfile(uploadedFullFileName, fullFileName);
            end
        end

        function esValido = validarEmail(email)
            valor = regexp(email,'[a-z0-9\_\.]+@[a-z0-9\.]+\.[a-z]{2,24}');
            if(valor == 1)
                esValido = true;
            else
                esValido = false;
            end
        end

        function esValido = validarContrasenyasIguales(contrasenya, repeatContrasenya)
            lCont = strlength(contrasenya);
            lRepeatCont = strlength(repeatContrasenya);
            if(lCont ~= lRepeatCont) 
                esValido = false;
            elseif(contrasenya ~= repeatContrasenya)
                esValido = false;
            else
                esValido = true;
            end
        end

        function tablaParseada = parsearCamposTablaMissingAEmpty(tablaAParsear)
            sizeTabla = size(tablaAParsear);
            for i = 1:sizeTabla(1)
                for j = 1:sizeTabla(2)
                    if(ismissing(tablaAParsear(i,j))) 
                        tablaAParsear(i,j) = {""};
                    end
                end                
            end
            tablaParseada = tablaAParsear;
        end  

        function uid = GenerarUUID()
            symbols = ['a':'z' 'A':'Z' '0':'9'];
            MIN_ST_LENGTH = 25;
            MAX_ST_LENGTH = 50;
            stLength = randi([MIN_ST_LENGTH,MAX_ST_LENGTH]);
            nums = randi(numel(symbols),[1 stLength]);
            uid = symbols (nums);
        end

        function nota = ObtenerNota(frecuencia)
            connSql = sqlite('.\bd\test.db');
            sql = "select nota from notas where ABS(frecuencia - " + frecuencia + ") = (select min (ABS (frecuencia - " + frecuencia + ")) from notas)";
            notaResult = fetch(connSql, sql);
            close(connSql);
            nota = notaResult.nota;
        end
    end
end